"""Receiver for seismological network datagrams

The protocol is intended to be able to describe accelerometrical activity in
general, which makes it very apt for recording and transmitting seismological
activity.

Module usage example:

.. code-block:: python

   import rattler

   for meas in rattler.measurements():
       print meas.timestamp, meas.values

Would output something like::

    2010-05-20 15:28:46.990220 (-0.03622, 0.05433, -0.97805)
    2010-05-20 15:28:47.005845 (-0.01811, 0.05433, -0.97805)
    2010-05-20 15:28:47.023423 (-0.01811, 0.05433, -0.97805)
    2010-05-20 15:28:47.041001 (-0.01811, 0.05433, -0.97805)
    2010-05-20 15:28:47.056626 (-0.01811, 0.05433, -0.97805)

But with much greater precision, as the numbers have been cut for brevity's
sake.
"""

# NOTE Remember to change version in setup.py as well.
__version__ = "1.0.2"

import socket
import struct
import datetime

class RattlerException(StandardError):
    pass

class PartialData(RattlerException):
    def __init__(self, num_got, num_want, data):
        self.num_got = num_got
        self.num_want = num_want
        self.data = data
        super(PartialData, self).__init__("wanted %d bytes, got %d"
                                          % (num_got, num_want))

class StructFormat(str):
    """Convenience object layer for procedural struct module."""
    def _get_size(self):
        return struct.calcsize(self)
    size = property(_get_size)

    def unpack_split(self, data):
        """Unpack own format from *data* and return (v, remainder)"""
        part, rest = data[:self.size], data[self.size:]
        return (struct.unpack(self, part), rest)

def hasattrs(obj, attrs):
    for attr in attrs:
        if not hasattr(obj, attr):
            return False
    return True

class Measurement(object):
    """A seismological measurement"""

    __slots__ = ("source", "timestamp", "x", "y", "z")

    def __init__(self, timestamp, x, y, z, source=None):
        self.timestamp = timestamp
        self.x, self.y, self.z = x, y, z
        self.source = source

    def clone(self, x, y, z):
        """Make a copy of this instance with new accelerometer data."""
        return type(self)(self.timestamp, x, y, z, source=self.source)

    def __str__(self):
        r = "<Measurement at %s" % (self.timestamp,)
        if self.source:
            r += " from %s" % (self.source,)
        r += ": (%+.6f, %+.6f, %+.6f)>" % (self.x, self.y, self.z)
        return r

    def __add__(self, o):
        if not hasattrs(o, ("x", "y", "z")):
            return NotImplemented
        return self.clone(self.x + o.x, self.y + o.y, self.z + o.z)
    def __sub__(self, o):
        if not hasattrs(o, ("x", "y", "z")):
            return NotImplemented
        return self.clone(self.x - o.x, self.y - o.y, self.z - o.z)

    def __mul__(self, n):
        return self.clone(self.x * n, self.y * n, self.z * n)
    def __div__(self, n):
        return self.clone(self.x / n, self.y / n, self.z / n)
    __truediv__ = __div__

    def __neg__(self):
        return self.clone(-self.x, -self.y, -self.z)
    def __pos__(self):
        return self.clone(+self.x, +self.y, +self.z)
    def __abs__(self):
        return self.clone(abs(self.x), abs(self.y), abs(self.z))
    def __invert__(self):
        return self.clone(~self.x, ~self.y, ~self.z)

    @property
    def values(self):
        return (self.x, self.y, self.z)

wrapper_fmt = StructFormat("! HH")
measurement_fmt = StructFormat("! f 3d")

def unpack_wrapper(msg):
    """Unpack the wrapper data for *msg*.

    The wrapper data consists of a message type and its size. This is returned
    as a two-tuple of (typ, data).

    If *msg* doesn't contain the whole message, a PartialData exception is
    raised.
    """
    (typ, sz), rest = wrapper_fmt.unpack_split(msg)
    if len(msg) != sz:
        raise PartialData(sz, len(rest), msg)
    return (typ, rest)

def unpack_measurement(msg):
    """Unpack the data of a measurement from *msg* into a four-tuple of
    (reltime, x, y, z).
    """
    rv, rest = measurement_fmt.unpack_split(msg)
    if rest:
        raise ValueError("got excess data %r" % (rest,))
    return rv

class SeismometerReceiver(object):
    msgtype_announce = (1 << 0)
    msgtype_measurement = (1 << 1)

    #: If True, measurements that are older than a previously seen measurement
    # will be ignored.
    drop_on_backward = True

    def __init__(self, ipv6=False, bind="0.0.0.0", port=5612):
        #: _first tracks what time difference there is between the seismometer
        # data and the local time. The seismometer's timestamps are based on
        # some arbitrary point in time.
        self._first = None
        self._prev_dt = None
        addrfam = socket.AF_INET
        if ipv6:
            addrfam = socket.AF_INET6
        self.sock = socket.socket(addrfam, socket.SOCK_DGRAM, 0)
        self.sock.bind((bind, port))

    def compensate_time(self, ts):
        if not self._first:
            (base_dt, base_ts) = (datetime.datetime.now(), ts)
            self._first = (base_dt, base_ts)
        else:
            (base_dt, base_ts) = self._first
        diff = ts - base_ts
        secs = int(diff)
        msecs = (diff % 1) * (10 ** 6)
        delta = datetime.timedelta(seconds=secs, microseconds=msecs)
        return base_dt + delta

    def receive(self):
        (data, addr) = self.sock.recvfrom(4096)
        typ, msg = unpack_wrapper(data)
        if typ == self.msgtype_announce:
            raise NotImplementedError("in the future!")
        elif typ == self.msgtype_measurement:
            (timestamp, x, y, z) = unpack_measurement(msg)
            timestamp = self.compensate_time(timestamp)
            if self.drop_on_backward:
                if self._prev_dt and timestamp < self._prev_dt:
                    return
                self._prev_dt = timestamp
            return Measurement(timestamp, x, y, z, source=addr)

def measurements():
    """Generator for measurement objects."""
    sr = SeismometerReceiver()
    while True:
        m = sr.receive()
        if m is None:
            continue
        yield m

if __name__ == "__main__":
    import sys

    outf = sys.stdout

    if sys.platform != "win32":
        outf.write("Awaiting initial measurement...")
        outf.flush()

    for meas in measurements():
        if sys.platform != "win32":
            # The ANSI control code 2K clears the current line.
            outf.write("\x1b[2K\r%s" % (meas,))  
        else:
            outf.write(str(meas) + "\n")
        outf.flush()



import sys
import datetime
from collections import deque
from math import sqrt

import rattler

# {{{ Utilities

def memorizer(iter, backlog=20):
    prev_vals = deque()
    for value in iter:
        prev_vals.append(value)
        while len(prev_vals) > backlog:
            prev_vals.popleft()
        yield list(prev_vals)

def medians(iter, backlog=20):
    for vals in memorizer(iter, backlog=backlog):
        vals.sort()
        mid_idx = len(vals) / 2
        if len(vals) & 1:
            yield vals[mid_idx]
        else:
            yield (vals[mid_idx - 1] + vals[mid_idx]) / 2.0

def ri(v): return int(round(v))
def dt2secs(dt):
    rv = (dt.days * 24 * 3600)
    rv += dt.seconds
    rv += dt.microseconds / 1000000.0
    return rv

# }}}

# {{{ pygame initialization

import pygame
from pygame import draw, display, rect
from pygame.locals import SWSURFACE, FULLSCREEN, HWSURFACE, DOUBLEBUF

pygame.init()
fullscreen = False

if fullscreen:
    depth = 0
    flags = FULLSCREEN | HWSURFACE | DOUBLEBUF
else:
    depth = 16
    flags = SWSURFACE | DOUBLEBUF

modes = display.list_modes(depth, flags)
if fullscreen:
    if modes == -1:  # Welcome to exceptionlessland
        raise SystemExit("Failed to initialize display")
    else:
        mode = max(modes)
else:
    mode = (640, 480)

display.set_mode(mode, flags)
screen = display.get_surface()
width, height = screen.get_size()

# }}}

fg = (0x30, 0x30, 0x30)
bg = (0xef, 0xef, 0xef)
cross_cl = (0xb0, 0xb0, 0xb0)

size = rect.Rect((0, 0), screen.get_size())
r1 = rect.Rect(size.topleft, size.center)
r2 = rect.Rect(size.midtop, size.center)
r3 = rect.Rect(size.midleft, size.center)
r4 = rect.Rect(size.center, size.center)
rs = (r1, r2, r3, r4)

views = ((r1, ("x", "y")),
         (r2, ("x", "z")),
         (r3, ("y", "z")),
         (r4, (None, None)))

# This is a simple 2D vector
class Vector(object):
    __slots__ = ("_x", "_y", "_length")

    def __init__(self, x, y):
        self._x, self._y = x, y
        self._length = sqrt(self.x ** 2 + self.y ** 2)

    @property
    def x(self): return self._x
    @property
    def y(self): return self._y
    @property
    def length(self): return self._length

    def normalized(self):
        return type(self)(self.x / self.length, self.y / self.length)

class LoopLatencies(object):
    clocksource = datetime.datetime.now

    def __init__(self):
        self.reset()

    def reset(self):
        self.start = self.clocksource()

    def __iter__(self):
        return self

    def next(self):
        dt = self.clocksource() - self.start
        latencies.reset()
        return dt2secs(dt)

latencies = LoopLatencies()
latency_medians = medians(latencies)

bg_r = r1.copy().move(0, 0)
line_len = min(bg_r.w, bg_r.h) / 2.0

bg_surf = pygame.Surface(bg_r.size, HWSURFACE)
bg_surf.fill(bg)
draw.circle(bg_surf, cross_cl, bg_r.center, ri(line_len), 1)
draw.line(bg_surf, cross_cl, (bg_r.left, bg_r.centery), (bg_r.right, bg_r.centery))
draw.line(bg_surf, cross_cl, (bg_r.centerx, bg_r.top), (bg_r.centerx, bg_r.bottom))
draw.rect(bg_surf, cross_cl, bg_r, 2)

n_skips = 0
for meas in rattler.measurements():
    lat = latency_medians.next()
    sys.stdout.write("\x1b[2K\rlatency = %.4fms" % (lat * 1000,))
    sys.stdout.flush()

    # Try to keep synchronized
    curr_ts = datetime.datetime.now()
    secs_ago = dt2secs(curr_ts - meas.timestamp)
    if secs_ago > (lat * 1.25):
        n_skips += 1
        sys.stdout.write(", skips = %d" % (n_skips,))
        sys.stdout.flush()
        continue
    else:
        n_skips = 0

    for (r, (va1, va2)) in views:
        if va1 is va2 is None:
            draw.rect(screen, bg, r)
            continue

        screen.blit(bg_surf, r.topleft)

        vec = Vector(getattr(meas, va1), getattr(meas, va2))

        if vec.length < 0.075:
            continue

        vec = vec.normalized()

        draw.line(screen, fg, r.center,
                  (ri(r.centerx + vec.x * line_len),
                   ri(r.centery + vec.y * line_len)))

    display.flip()



import sys
from collections import deque

import signal

def sigint(frame, signo):
    sys.stdout.write("\x1b[2K\rBye!\n")
    sys.exit(0)

signal.signal(signal.SIGINT, sigint)

def medians(iter, backlog=20):
    """Given an iterable, yields median values from the *backlog* number of
    last items.
    
    Every value taken from *iter* must be comparable to every other value from
    that place.
    """
    prev_vals = deque()
    for value in iter:
        prev_vals.append(value)
        while len(prev_vals) > backlog:
            prev_vals.popleft()
        # Median part, could be split out
        vals = list(prev_vals)
        vals.sort()
        mid_idx = len(vals) / 2
        if len(vals) & 1:
            yield vals[mid_idx]
        else:
            yield (vals[mid_idx - 1] + vals[mid_idx]) / 2.0

import rattler

for meas in medians(rattler.measurements()):
    adjs = []
    p = adjs.append

    # Use Z to determine facing direction
    if meas.z > 0.8:
        p("flat, display down")
    elif meas.z > 0.3:
        p("tilted, display down")
    elif meas.z < -0.8:
        p("flat, display up")
    elif meas.z < -0.3:
        p("tilted, display up")

    # Use X & Y to determine which side is down
    xy_tilt = ((abs(meas.x) + abs(meas.y)) / 2)
    if xy_tilt <= 0.2:
        p("lying down")
    elif meas.y < -0.8:
        p("standing")
    elif meas.y > 0.8:
        p("standing upside down")
    elif meas.x < -0.8:
        p("lying on left side")
    elif meas.x > 0.8:
        p("lying on right side")

    desc = ", ".join(adjs)
    sys.stdout.write("\x1b[2K\r" + desc)
    sys.stdout.flush()



from distutils.core import setup

# NOTE Remember to change version in rattler.py as well.
setup(name="rattler", version="1.0.2",
      author="Yellow Agents AB",
      author_email="ludvig@yellowagents.com",
      url="http://yellowagents.com/seismometer/",
      py_modules=["rattler"])
